import hashlib
import itertools
import multiprocessing
import os
import time

import backoff as backoff
import requests
from diskcache import FanoutCache

from .compressed_disk import CompressedDisk
from simtools.Analysis.DataRetrievalProcess import retrieve_SSMT_files, retrieve_COMPS_AM_files
from simtools.Analysis.OutputParser import SimulationOutputParser

file_cache = FanoutCache('/home/ccollins/.comps_file_cache', shards=multiprocessing.cpu_count(), timeout=120,
                         disk=CompressedDisk, disk_compress_level=5,
                         size_limit=2 ** 40)


@backoff.on_exception(backoff.expo, requests.exceptions.RequestException, max_time=300)
def retrieve_data_for_simulation_aggregate(q, simulation, aggregators, path_mapping):
    # Filter first and get the filenames from filtered analysis
    filtered_aggregator = [a for a in aggregators if a.filter(simulation)]
    filenames = set(itertools.chain(*(a.filenames for a in filtered_aggregator)))

    # We dont have anything to do :)
    if not filtered_aggregator:
        return {}

    # The byte_arrays will associate filename with content
    byte_arrays = {}

    # Retrieval for SSMT
    if path_mapping:
        byte_arrays = retrieve_SSMT_files(simulation, filenames, path_mapping)

    # Retrieval for normal HPC Asset Management
    elif simulation.experiment.location == "HPC":
        # global comps cache for local runs

        m = hashlib.md5()
        name = f'{simulation.id}_{"_".join(list(sorted(filenames)))}'
        m.update(name.encode('utf-8'))
        name = m.hexdigest()
        if name not in file_cache:
            # print(f'Retrieving from HPC fpr {name}')
            byte_arrays = retrieve_COMPS_AM_files(simulation, filenames)
            file_cache[name] = byte_arrays
        else:
            # print('Using cache')
            byte_arrays = file_cache[name]

    # Retrieval for local file
    else:
        for filename in filenames:
            path = os.path.join(simulation.get_path(), filename)
            with open(path, 'rb') as output_file:
                byte_arrays[filename] = output_file.read()

    # Selected data will be a dict with analyzer.uid => data
    selected_data = {}
    for aggregator in filtered_aggregator:
        # If the analyzer needs the parsed data, parse
        if aggregator.parse:
            data = {filename: SimulationOutputParser.parse(filename, content)
                    for filename, content in byte_arrays.items()}
        else:
            data = byte_arrays

        # Retrieve the selected data for the given analyzer
        selected_data[aggregator.uid] = aggregator.select_simulation_data(data, simulation)
    # wait until the queue has room
    while q.full():
        time.sleep(5)
    q.put(selected_data)
