# From https://comps.idmod.org/#explore/WorkItems?filters=Id=b5359d6a-9ca1-e911-a2c1-c4346bcb1555&offset=0&related=true&count=10&selectedId=b5359d6a-9ca1-e911-a2c1-c4346bcb1555
import os
import pandas as pd
import numpy as np
import shutil

from simtools.Analysis.BaseAnalyzers import BaseAnalyzer
from simtools.Analysis.AnalyzeManager import AnalyzeManager

projectdir = os.path.join(os.path.expanduser('~'), 'Dropbox (IDM)', 'Malaria Team Folder', 'projects',
                          'Vector_genetics', 'Data', 'simulation_data', 'Establishment_Burkina')


def construct_spatial_output_df(rawdata, channel, timesteps=[]):
    n_nodes = rawdata['n_nodes']
    n_tstep = rawdata['n_tstep']
    if 'start' in rawdata:
        start = rawdata['start']
        interval = rawdata['interval']
    else:
        start, interval = 0, 1
    nodeids = rawdata['nodeids']
    data = rawdata['data']

    all_timesteps = range(start, (start + n_tstep) * interval, interval)

    df = pd.DataFrame({channel: [item for sublist in data for item in sublist],
                       'time': [item for sublist in [[x] * n_nodes for x in all_timesteps] for item in sublist],
                       'node': [item for sublist in [list(nodeids) * len(all_timesteps)] for item in sublist]})
    if not timesteps:
        return df

    timesteps = sorted(list(set(all_timesteps).intersection(timesteps)))
    return df[df['time'].isin(timesteps)]


class GMEstablishmentAnalyzer(BaseAnalyzer):

    def __init__(self, exp_name, working_dir='.'):
        super(GMEstablishmentAnalyzer, self).__init__(working_dir=working_dir,
                                                      filenames=['output/'
                                                                 'ReportVectorGenetics_gambiae_Female_SPECIFIC_GENOME.csv']
                                                      )

        self.exp_name = exp_name
        self.sweep_variables = ['Run_Number', 'Release_number']
        self.output_fname = os.path.join(self.working_dir, "%s_establishment_rates.csv" % self.exp_name)

    def select_simulation_data(self, data, simulation):

        simdata = data['output/ReportVectorGenetics_gambiae_Female_SPECIFIC_GENOME.csv']
        genomes = simdata['Genome'].unique()
        simdata = simdata.pivot_table('VectorPopulation', ['Time', 'NodeID'], 'Genome').reset_index()
        simdata['sum'] = [0] * len(simdata)
        for genome in genomes:
            simdata['sum'] += simdata[genome]
        for genome in genomes:
            simdata[genome] = simdata[genome] / simdata['sum']
            simdata = simdata.rename(columns={'X-a0:X-a0': 'WT'})
        genomes = genomes.tolist()
        genomes.remove('X-a0:X-a0')
        simdata['GM'] = [0] * len(simdata)
        for genome in genomes:
            simdata['GM'] += simdata[genome]
        simdata = simdata[['Time', 'NodeID', 'GM', 'WT']]

        for sweep_var in self.sweep_variables:
            if sweep_var in simulation.tags.keys():
                simdata[sweep_var] = simulation.tags[sweep_var]
            else:
                simdata[sweep_var] = 0
        return simdata

    def finalize(self, all_data):
        data_sets_per_experiment = {}

        for simulation, associated_data in all_data.items():
            experiment_name = simulation.experiment.exp_name
            if experiment_name not in data_sets_per_experiment:
                data_sets_per_experiment[experiment_name] = []

            data_sets_per_experiment[experiment_name].append(associated_data)

        for experiment_name, data_sets in data_sets_per_experiment.items():
            d = pd.concat(data_sets).reset_index(drop=True)
            d = d.groupby('Run_Number')['GM'].apply(np.mean).reset_index()
            d.to_csv(self.output_fname, index=False)


class SpatialAnalyzer(BaseAnalyzer):

    def __init__(self, exp_name, spatial_channels, working_dir='.'):
        super(SpatialAnalyzer, self).__init__(working_dir=working_dir,
                                              filenames=['output/SpatialReportMalariaFiltered_%s.bin' % x for x in
                                                         spatial_channels]
                                              )

        self.exp_name = exp_name
        self.sweep_variables = ['Run_Number', 'Release_number']
        self.spatial_channels = spatial_channels
        self.output_fname = os.path.join(self.working_dir, "%s_spatial_data.csv" % self.exp_name)

    def select_simulation_data(self, data, simulation):

        simdata = construct_spatial_output_df(
            data['output/SpatialReportMalariaFiltered_%s.bin' % self.spatial_channels[0]], self.spatial_channels[0])
        if len(self.spatial_channels) > 1:
            for ch in self.spatial_channels[1:]:
                simdata = pd.merge(left=simdata,
                                   right=construct_spatial_output_df(
                                       data['output/SpatialReportMalariaFiltered_%s.bin' % ch], ch),
                                   on=['time', 'node'])

        for sweep_var in self.sweep_variables:
            if sweep_var in simulation.tags.keys():
                simdata[sweep_var] = simulation.tags[sweep_var]
            else:
                simdata[sweep_var] = 0
        # return simdata
        # ensure data directory
        # because we are running in threads, this could happen in multiple threads
        try:
            os.makedirs('intermediate_data', exist_ok=True)
        except FileExistsError:
            pass
        out_path = f'intermediate_data/{simulation.id}.pickle'

        simdata.to_pickle(out_path)
        # return out_path
        return out_path

    def finalize(self, all_data):
        # load all the simulations
        all_data_dict = dict()
        for simulation, path in all_data.items():
            all_data_dict[simulation] = pd.read_pickle(path)

        data_sets_per_experiment = {}

        # for simulation, associated_data in all_data.items():
        for simulation, associated_data in all_data_dict.items():
            experiment_name = simulation.experiment.exp_name
            if experiment_name not in data_sets_per_experiment:
                data_sets_per_experiment[experiment_name] = []

            data_sets_per_experiment[experiment_name].append(associated_data)

        for experiment_name, data_sets in data_sets_per_experiment.items():
            d = pd.concat(data_sets).reset_index(drop=True)
            d = d.groupby('Run_Number')[self.spatial_channels].apply(np.mean).reset_index()
            d.to_csv(self.output_fname, index=False)


if __name__ == "__main__":

    experiments = {
        # "VC_only": "36f37682-158a-e911-a2c1-c4346bcb1555",
        # "VC_and_GM": "35f37682-158a-e911-a2c1-c4346bcb1555",
        # "No_interventions": "e264b5ee-be8b-e911-a2c1-c4346bcb1555",
        # "GM_only": "4880820c-8991-e911-a2c1-c4346bcb1555",
        "VC_and_GM_vary_release": "b8bb4828-a599-e911-a2c1-c4346bcb1555"
    }
    channels = ['Prevalence', 'Daily_EIR', 'Adult_Vectors']

    for expt_name, exp_id in experiments.items():
        am = AnalyzeManager(exp_list=exp_id,
                            analyzers=
                            [SpatialAnalyzer(exp_name=expt_name,
                                             spatial_channels=
                                             ['Prevalence', 'Daily_EIR', 'Adult_Vectors',
                                              'New_Clinical_Cases']
                                             ),
                             GMEstablishmentAnalyzer(exp_name=expt_name,
                                                     )
                             ]
                            )
        print(am.experiments)
        am.analyze()
    try:
        shutil.rmtree('./intermediate_data')
    except Exception as e:
        print(str(e))
        pass
