from abc import abstractmethod

import collections
import re
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import List

import pandas as pd
from COMPS.Data.Simulation import SimulationState
from tqdm import tqdm
import itertools
import multiprocessing
import os
import time

import backoff as backoff
import requests

from simtools.Analysis.DataRetrievalProcess import retrieve_SSMT_files, retrieve_COMPS_AM_files
from simtools.Analysis.OutputParser import SimulationOutputParser
from simtools.DataAccess.Schema import Experiment
from simtools.Utilities.COMPSCache import COMPSCache
from simtools.Utilities.COMPSUtilities import COMPS_login
from simtools.Utilities.Experiments import retrieve_experiment


def monitor_queue(queue, total, aggregators, filename='aggregated_data_debu.h5'):
    print("Starting aggregator process")
    print([aggregator.uid for aggregator in aggregators])
    aggregate_output = pd.HDFStore(filename, mode='w', compression='blosc', complevel=5)

    processed = 0
    with tqdm(total=total) as pbar:
        while True:
            data = queue.get()
            #print(f'{processed}')
            for aggregator in aggregators:
                if aggregator.uid in data:
                    pbar.set_description(f"Appending {processed} data to {aggregator.uid}")
                    aggregate_output.append(aggregator.uid, data[aggregator.uid])
            pbar.update()
            processed += 1
            if processed >= total:
                break


@backoff.on_exception(backoff.expo, requests.exceptions.RequestException, max_time=300)
def retrieve_data_for_simulation_aggregate(q, simulation, aggregators, path_mapping):
    # Filter first and get the filenames from filtered analysis
    filtered_aggregator = [a for a in aggregators if a.filter(simulation)]
    filenames = set(itertools.chain(*(a.filenames for a in filtered_aggregator)))

    # We dont have anything to do :)
    if not filtered_aggregator:
        q.put({})

    # The byte_arrays will associate filename with content
    byte_arrays = {}

    # Retrieval for SSMT
    if path_mapping:
        byte_arrays = retrieve_SSMT_files(simulation, filenames, path_mapping)

    # Retrieval for normal HPC Asset Management
    elif simulation.experiment.location == "HPC":
        # global comps cache for local runs

        byte_arrays = retrieve_COMPS_AM_files(simulation, filenames)

    # Retrieval for local file
    else:
        for filename in filenames:
            path = os.path.join(simulation.get_path(), filename)
            with open(path, 'rb') as output_file:
                byte_arrays[filename] = output_file.read()

    # Selected data will be a dict with analyzer.uid => data
    selected_data = {}
    for aggregator in filtered_aggregator:
        # If the analyzer needs the parsed data, parse
        if aggregator.parse:
            data = {filename: SimulationOutputParser.parse(filename, content)
                    for filename, content in byte_arrays.items()}
        else:
            data = byte_arrays

        # Retrieve the selected data for the given analyzer
        selected_data[aggregator.uid] = aggregator.select_simulation_data(data, simulation)
    # wait until the queue has room
    while q.full():
        time.sleep(5)
    q.put(selected_data)


class BaseAggregator:
    @abstractmethod
    def __init__(self, need_dir_map=False, uid=None, working_dir=None, filenames=None, parse=True):
        self.filenames = filenames or []
        self.need_dir_map = need_dir_map
        self.working_dir = working_dir
        self.parse = parse
        self.uid = uid or self.__class__.__name__

    def filter(self, simulation):
        """
        Decide whether analyzer should process a simulation
        :param simulation: simulation object
        :return: Boolean whether simulation should be analyzed by this analyzer
        """
        return True

    def select_simulation_data(self, data, simulation):
        """
        In parallel for each simulation, consume raw data from filenames and emit selected data
        :param data: simulation data. Dictionary associating filename with content
        :param simulation: object representing the simulation for which the data is passed
        :return: selected data for the given simulation
        """
        return {}


class AggregateDataManager:
    def __init__(self, aggregators: List[BaseAggregator], exp_list, max_sims=None, working_dir=None,
                 queue_function=monitor_queue):
        self.aggregators: List[BaseAggregator] = [aggregators] if type(aggregators) is BaseAggregator else aggregators
        self.experiments = set()
        self.max_sims = max_sims
        self.ignored_simulations = {}
        self.simulations = {}
        self.working_dir = working_dir or os.getcwd()
        self.verbose = True
        # Check if we are on SSMT
        ssmt_path_mapping = os.environ.get("COMPS_DATA_MAPPING", None)
        if ssmt_path_mapping:
            ssmt_path_mapping.lower().split(';')

        # Initial adding of experiments
        if exp_list:
            exp_list = exp_list if isinstance(exp_list, collections.Iterable) and not isinstance(exp_list,
                                                                                                 str) else [
                exp_list]
            for exp in exp_list: self.add_experiment(exp)

    def add_experiment(self, experiment: Experiment):
        if not isinstance(experiment, Experiment):
            experiment = retrieve_experiment(experiment)

        if experiment not in self.experiments:
            self.experiments.add(experiment)
            if experiment.location == "HPC":
                COMPS_login(experiment.endpoint)
                COMPSCache.load_experiment(experiment.exp_id)

            self.filter_simulations(experiment.simulations)

    def filter_simulations(self, simulations):
        if self.max_sims is not None:
            # If we already reached the maximum_simulation count -> exit
            if self.max_sims <= 0: return

            # Truncate the simulations to only add the needed ones
            simulations = simulations[:self.max_sims - len(self.simulations)]
            self.max_sims -= len(self.simulations)

        for s in simulations:
            if s.status != SimulationState.Succeeded:
                self.ignored_simulations[s.id] = s
            else:
                self.simulations[s.id] = s

    def add_aggregator(self, aggregator):
        # First check if we need to change the UID depending on other analyzers
        number = 1
        while any(a.uid == aggregator.uid for a in self.aggregators):
            # If we already suffixed with a number, remove it
            aggregator.uid = re.sub("\(\d+\)$", "", aggregator.uid)
            # Add the suffix
            aggregator.uid += "({})".format(number)
            number += 1

        aggregator.working_dir = aggregator.working_dir or self.working_dir

        # Then call the initialize method
        aggregator.initialize()

        self.aggregators.append(aggregator)

    def analyze(self):
        # Start the timer
        start_time = time.time()

        # If no analyzers -> quit
        if not all((self.aggregators, self.simulations)):
            print("No analyzers or experiments selected, exiting...")
            return False

        # Check if we are on SSMT
        ssmt_path_mapping = os.environ.get("COMPS_DATA_MAPPING", None)
        if ssmt_path_mapping:
            ssmt_path_mapping.lower().split(';')

        # If any of the analyzer needs the dir map, create it
        # Or if we are on SSMT
        if ssmt_path_mapping or any(a.need_dir_map for a in self.aggregators):
            # preload the global dir map
            from simtools.Utilities.SimulationDirectoryMap import SimulationDirectoryMap
            for experiment in self.experiments:
                SimulationDirectoryMap.preload_experiment(experiment)

        scount = len(self.simulations)

        if scount == 0 and self.verbose:
            print("No experiments/simulations for analysis.")
            return False

        with multiprocessing.Manager() as manager:
            q = manager.Queue(maxsize=64)
            write_process = multiprocessing.Process(target=monitor_queue, args=(q, scount, self.aggregators))
            write_process.start()
            with ProcessPoolExecutor(max_workers=multiprocessing.cpu_count()) as executor:
                futures = []
                for simulation in tqdm(self.simulations.values()):
                    futures.append(
                        executor.submit(retrieve_data_for_simulation_aggregate, q, simulation, self.aggregators,
                                        ssmt_path_mapping))

                for f in as_completed(futures):
                    f.result()

            # nearly done. we need to check queue and monitor until it is empty
            while not q.empty():
                time.sleep(5)
            write_process.join()
        return True
