import numpy as np
import math
import os
import json
import pandas as pd
import configparser

from malaria.study_sites.site_setup_functions import summary_report_fn
from dtk.interventions.mosquito_release import add_mosquito_release
from dtk.interventions.itn import add_ITN
from dtk.vector.species import set_species_trait_modifiers, set_species_drivers


# from createSimDirectoryMapBR import createSimDirectoryMap


def add_summary_report(cb, start_day=0, ipfilter='', description=''):
    summary_report_fn(start=start_day + 1, interval=30.0, description='Monthly_Report_%s' % description,
                      age_bins=[5.0, 15.0, 100.0],
                      parasitemia_bins=[0.0, 0.01, 0.1, 4.0, 40.0, 4000000.0],
                      infection_bins=np.arange(0.0, 101.0, 5.0).tolist(),
                      ipfilter=ipfilter,
                      nodes={'Node_List': [i for i in range(1, 89)], "class": "NodeSetNodeList"})(cb)

    return None


def get_outpath_for_serialized_file(simmap, seed=0):
    temp = simmap[simmap['Run_Number'] == seed]

    return temp[temp['Run_Number'] == seed]['outpath'].values[0]


def make_simmap(expname):
    simmap = createSimDirectoryMap(expname)

    return simmap


def add_release(cb, number=100, num_nodes=1, start_day=180):
    nodelist = find_n_largest(n=num_nodes)
    add_mosquito_release(cb, start_day=start_day, species='gambiae', repetitions=1, number=100,
                         released_genome=[['X', 'Y'], ['a1', 'a1']],
                         nodes={'Node_List': nodelist, "class": "NodeSetNodeList"})

    return {'Release_number': number}


def add_nets(cb, coverage=0.6):
    VC_coverage = {'min': 0, 'max': 100, 'coverage': coverage}

    # ITN
    add_ITN(cb, 0, [VC_coverage])
    add_ITN(cb, 3 * 365, [VC_coverage])

    return {'ITN_Coverage': coverage}


def add_trait_modifiers(cb, transmission_to_human=0.0):
    # Add gene trait modifiers
    traits = {'gambiae': [
        {
            "Allele_Combinations": [["a0", "a1"]],
            "Trait_Modifiers": {"TRANSMISSION_TO_HUMAN": transmission_to_human}
        },
        {
            "Allele_Combinations": [["a1", "a1"]],
            "Trait_Modifiers": {"TRANSMISSION_TO_HUMAN": transmission_to_human}
        }
    ]
    }

    set_species_trait_modifiers(cb, traits)

    return {'Transmission_To_Human': transmission_to_human}


def add_drivers(cb, copy_to_likelihood=1.0):
    drivers = {'gambiae': [{
        "Alleles_Driven": [
            {"Allele_To_Copy": "a1",
             "Allele_To_Replace": "a0",
             "Copy_To_Likelihood": {"a0": 1 - copy_to_likelihood, "a1": copy_to_likelihood}
             }],
        "Driver_Type": "CLASSIC",
        "Driving_Allele": "a1"
    }]
    }

    set_species_drivers(cb, drivers)

    return {'Copy_To_Likelihood': copy_to_likelihood}


# Input: intervention_df = pd.read_csv("grid_itn_events.csv")
# format: interventions by grid cell.  Has columns for "grid_cell", "fulldate", "simday"
def try_campaign_compression(intervention_df, field_list, bin_fidelity=0.1, simday_fidelity=10):
    def round_nearest(x, a):
        rounded = round(round(x / a) * a, -int(math.floor(math.log10(a))))
        return rounded

    binned_intervention_df = intervention_df.copy(deep=True)

    # Bin these fields with bin width = bin_fidelity
    for field in field_list:
        binned_intervention_df[field] = intervention_df[field].map(lambda x: round_nearest(x, bin_fidelity))

    filter_on_positive_columns = ['cov_all', 'coverage', 'trigger_coverage']
    for col in [x for x in filter_on_positive_columns if x in binned_intervention_df.columns]:
        binned_intervention_df = binned_intervention_df[binned_intervention_df[col] > 0]

    if 'simday' in intervention_df.columns:
        binned_intervention_df['simday'] = intervention_df['simday'].map(
            lambda x: round_nearest(x, simday_fidelity)).astype(int)
        # Group by the new binned fields, as well as by date
        data_fields = ['simday'] + field_list
    else:
        data_fields = field_list
    binned_and_grouped = binned_intervention_df.groupby(data_fields)

    return [binned_intervention_df, binned_and_grouped, data_fields]


def find_n_largest(n=1):
    config = configparser.ConfigParser()
    config.read('simtools.ini')
    direc = config['HPC']['input_root']

    file = os.path.join(direc, 'Demographics', '%s_demographics.json' % 'vector_genetics')
    nodes = []
    pop = []
    with open(file) as f:
        data = json.load(f)
        for node in data['Nodes']:
            nodes.append(node['NodeID'])
            pop.append(node['NodeAttributes']['InitialPopulation'])
    df = pd.DataFrame({'nodes': nodes, 'pop': pop})
    df = df.sort_values(by=['pop'], ascending=False)

    node_list = list(df['nodes'][:n])

    return node_list


column_types = dict(Time='uint16',
                    time='uint16',
                    New_Clinical_Cases='uint8',
                    Adult_Vectors='uint32',
                    Run_Number='uint16',
                    Release_number='uint16',
                    Node='uint8',
                    node='uint8')


def optimize_dataframe(simdata):
    """
    Set the data to the optimal size for their types. For examples, we know this experiment does not have
    more than 255 nodes so we should assign the NodeId columns to an uint8 since it takes less memory than the
    default uint64

    :param simdata:
    :return:
    """
    for column, item_type in column_types.items():
        if column in simdata.columns:
            simdata[column] = simdata[column].astype(item_type)
    return simdata
